#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .GXDLMSQualityOfService import GXDLMSQualityOfService

# pylint: disable=too-many-instance-attributes
class GXDLMSGprsSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSGprsSetup
    """

    def __init__(self, ln="0.0.25.4.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.GPRS_SETUP, ln, sn)
        self.apn = ""
        self.pinCode = 0
        self.defaultQualityOfService = GXDLMSQualityOfService()
        self.requestedQualityOfService = GXDLMSQualityOfService()

    def getValues(self):
        return [self.logicalName,
                self.apn,
                self.pinCode,
                self.defaultQualityOfService,
                self.requestedQualityOfService]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  APN
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  PINCode
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  DefaultQualityOfService + RequestedQualityOfService
        if all_ or not self.isRead(4):
            attributes.append(4)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 4

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.OCTET_STRING
        elif index == 3:
            ret = DataType.UINT16
        elif index == 4:
            ret = DataType.ARRAY
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            if not self.apn:
                ret = None
            else:
                ret = self.apn.encode()
        elif e.index == 3:
            ret = self.pinCode
        elif e.index == 4:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(5)
            _GXCommon.setData(settings, data, DataType.UINT8, self.defaultQualityOfService.precedence)
            _GXCommon.setData(settings, data, DataType.UINT8, self.defaultQualityOfService.delay)
            _GXCommon.setData(settings, data, DataType.UINT8, self.defaultQualityOfService.reliability)
            _GXCommon.setData(settings, data, DataType.UINT8, self.defaultQualityOfService.peakThroughput)
            _GXCommon.setData(settings, data, DataType.UINT8, self.defaultQualityOfService.meanThroughput)
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(5)
            _GXCommon.setData(settings, data, DataType.UINT8, self.requestedQualityOfService.precedence)
            _GXCommon.setData(settings, data, DataType.UINT8, self.requestedQualityOfService.delay)
            _GXCommon.setData(settings, data, DataType.UINT8, self.requestedQualityOfService.reliability)
            _GXCommon.setData(settings, data, DataType.UINT8, self.requestedQualityOfService.peakThroughput)
            _GXCommon.setData(settings, data, DataType.UINT8, self.requestedQualityOfService.meanThroughput)
            ret = data.array()
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if isinstance(e.value, str):
                self.apn = e.value
            else:
                self.apn = _GXCommon.changeType(settings, e.value, DataType.STRING)
        elif e.index == 3:
            self.pinCode = e.value
        elif e.index == 4:
            arr = (e.value)[0]
            self.defaultQualityOfService.precedence = arr[0]
            self.defaultQualityOfService.delay = arr[1]
            self.defaultQualityOfService.reliability = arr[2]
            self.defaultQualityOfService.peakThroughput = arr[3]
            self.defaultQualityOfService.meanThroughput = arr[4]
            arr = (e.value)[1]
            self.requestedQualityOfService.precedence = arr[0]
            self.requestedQualityOfService.delay = arr[1]
            self.requestedQualityOfService.reliability = arr[2]
            self.requestedQualityOfService.peakThroughput = arr[3]
            self.requestedQualityOfService.meanThroughput = arr[4]
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.apn = reader.readElementContentAsString("APN")
        self.pinCode = reader.readElementContentAsInt("PINCode")
        if reader.isStartElement("DefaultQualityOfService", True):
            self.defaultQualityOfService.precedence = reader.readElementContentAsInt("Precedence")
            self.defaultQualityOfService.delay = reader.readElementContentAsInt("Delay")
            self.defaultQualityOfService.reliability = reader.readElementContentAsInt("Reliability")
            self.defaultQualityOfService.peakThroughput = reader.readElementContentAsInt("PeakThroughput")
            self.defaultQualityOfService.meanThroughput = reader.readElementContentAsInt("MeanThroughput")
            reader.readEndElement("DefaultQualityOfService")
        if reader.isStartElement("RequestedQualityOfService", True):
            self.requestedQualityOfService.precedence = reader.readElementContentAsInt("Precedence")
            self.requestedQualityOfService.delay = reader.readElementContentAsInt("Delay")
            self.requestedQualityOfService.reliability = reader.readElementContentAsInt("Reliability")
            self.requestedQualityOfService.peakThroughput = reader.readElementContentAsInt("PeakThroughput")
            self.requestedQualityOfService.meanThroughput = reader.readElementContentAsInt("MeanThroughput")
            reader.readEndElement("DefaultQualityOfService")

    def save(self, writer):
        writer.writeElementString("APN", self.apn)
        writer.writeElementString("PINCode", self.pinCode)
        if self.defaultQualityOfService:
            writer.writeStartElement("DefaultQualityOfService")
            writer.writeElementString("Precedence", self.defaultQualityOfService.precedence)
            writer.writeElementString("Delay", self.defaultQualityOfService.delay)
            writer.writeElementString("Reliability", self.defaultQualityOfService.reliability)
            writer.writeElementString("PeakThroughput", self.defaultQualityOfService.peakThroughput)
            writer.writeElementString("MeanThroughput", self.defaultQualityOfService.meanThroughput)
            writer.writeEndElement()
        if self.requestedQualityOfService:
            writer.writeStartElement("RequestedQualityOfService")
            writer.writeElementString("Precedence", self.requestedQualityOfService.precedence)
            writer.writeElementString("Delay", self.requestedQualityOfService.delay)
            writer.writeElementString("Reliability", self.requestedQualityOfService.reliability)
            writer.writeElementString("PeakThroughput", self.requestedQualityOfService.peakThroughput)
            writer.writeElementString("MeanThroughput", self.requestedQualityOfService.meanThroughput)
            writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums.DataProtectionKeyType import DataProtectionKeyType
from .GXDLMSDataProtectionIdentifiedKey import GXDLMSDataProtectionIdentifiedKey
from .GXDLMSDataProtectionWrappeddKey import GXDLMSDataProtectionWrappeddKey
from .GXDLMSDataProtectionAgreedKey import GXDLMSDataProtectionAgreedKey


class GXDLMSDataProtectionKey:
    """
    Data protection Key.
    """

    def __init__(self):
        """
        Constructor.
        """

        self.__identifiedKey = GXDLMSDataProtectionIdentifiedKey()
        self.__wrappedKey = GXDLMSDataProtectionWrappeddKey()
        self.__agreedKey = GXDLMSDataProtectionAgreedKey()
        self.__dataProtectionKeyType = DataProtectionKeyType.IDENTIFIED

    @property
    def dataProtectionKeyType(self):
        """
        Data protection key type.
        """
        return self.__dataProtectionKeyType

    @dataProtectionKeyType.setter
    def dataProtectionKeyType(self, value):
        self.__dataProtectionKeyType = value

    @property
    def identifiedKey(self):
        """
        Identified key parameters.
        """
        return self.__identifiedKey

    @identifiedKey.setter
    def identifiedKey(self, value):
        self.__identifiedKey = value

    @property
    def wrappedKey(self):
        """
        Wrapped key parameters.
        """
        return self.__wrappedKey

    @wrappedKey.setter
    def wrappedKey(self, value):
        self.__wrappedKey = value

    @property
    def agreedKey(self):
        """
        Agreed key parameters.
        """
        return self.__agreedKey

    @agreedKey.setter
    def agreedKey(self, value):
        self.__agreedKey = value



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXByteBuffer import GXByteBuffer


# Long get or set information is saved here.
#
class GXDLMSLongTransaction:
    # pylint: disable=too-few-public-methods
    #
    # Constructor.
    #
    # @param forTargets
    #            Targets.
    # @param forCommand
    #            Command.
    # @param forData
    #            Data.
    #
    def __init__(self, forTargets, forCommand, forData=None):
        # Targets.
        self.targets = forTargets
        # Executed command.
        self.command = forCommand
        # Extra data from PDU.
        self.data = GXByteBuffer()
        if forData:
            self.data.set(forData)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename:        $HeadURL$
#
#  Version:         $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import xml.etree.cElementTree as ET
from .secure.GXDLMSSecureClient import GXDLMSSecureClient
from .enums import InterfaceType, Command, DataType
from .GXDLMS import GXDLMS
from .internal._GXCommon import _GXCommon
from .GXByteBuffer import GXByteBuffer
from .GXDLMSLNParameters import GXDLMSLNParameters
from .GXDLMSXmlSettings import GXDLMSXmlSettings
from .GXDLMSTranslator import GXDLMSTranslator
from .enums.Security import Security
from .GXDLMSXmlPdu import GXDLMSXmlPdu
from .enums.TranslatorOutputType import TranslatorOutputType

class GXDLMSXmlClient(GXDLMSSecureClient):
    """
    GXDLMS Xml client implements methods to communicate with DLMS/COSEM metering
    devices using XML.
    """

    #
    # Constructor
    #
    # @param type
    #            XML type.
    #
    def __init__(self, type_ = TranslatorOutputType.SIMPLE_XML):
        GXDLMSSecureClient.__init__(self)
        # XML client don't throw exceptions. It serializes them as a default. Set
        # value to true, if exceptions are thrown.
        #
        self.throwExceptions = False
        self.translator = GXDLMSTranslator(type_)
        self.translator.hex = False
        self.useLogicalNameReferencing = True


    @classmethod
    def removeRecursively(cls, node, nodeType, name):
        if node.getNodeType() == nodeType and (name is None or node.tag == name):
            node.getParentNode().removeChild(node)
        else:
            #  check the children recursively
            list_ = node.getChildNodes()
            i = 0
            while i < list_.getLength():
                cls.removeRecursively(list_.item(i), nodeType, name)
                i += 1

    #
    # Load XML commands from the string or file.
    #
    # @param filename
    #            XML file name.
    # @param settings
    #            Load settings.
    # Loaded XML objects.
    # pylint: disable=too-many-locals,too-many-nested-blocks
    def load(self, filename, loadSettings = None):
        tree = ET.parse(filename)
        root = tree.getroot()
        actions = []
        description = None
        error = None
        errorUrl = None
        sleep = None
        for node in root:
            if node.tag == "Description":
                description = node.text
                continue
            if node.tag == "Error":
                error = node.text
                continue
            if node.tag == "ErrorUrl":
                errorUrl = node.text
                continue
            if node.tag == "Sleep":
                sleep = node.text
                continue
            if loadSettings and node.tag == "GetRequest":
                structure = node.find("./GetRequestNormal/AccessSelection/AccessSelector/AccessParameters/Structure")
                if structure:
                    start = False
                    for node2 in structure:
                        if start:
                            bb = GXByteBuffer()
                            if start:
                                _GXCommon.setData(self.settings, bb, DataType.OCTET_STRING, loadSettings.start)
                                node2.attrib["Value"] = bb.toHex(False, 2)
                                start = False
                            else:
                                _GXCommon.setData(self.settings, bb, DataType.OCTET_STRING, loadSettings.end)
                                node2.attrib["Value"] = bb.toHex(False, 2)

            s = GXDLMSXmlSettings(self.translator.outputType, self.translator.hex, self.translator.showStringAsHex, self.translator.tagsByName)
            s.settings.clientAddress = self.settings.clientAddress
            s.settings.serverAddress = self.settings.serverAddress
            reply = []
            reply = self.translator.xmlToPdu(GXDLMSXmlPdu.getOuterXml(node), s)
            if s.command == Command.SNRM and not s.settings.isServer:
                self.settings.hdlc.maxInfoTX = s.settings.hdlc.maxInfoTX
                self.settings.hdlc.maxInfoRX = s.settings.hdlc.maxInfoRX
                self.settings.hdlc.windowSizeRX = s.settings.hdlc.windowSizeRX
                self.settings.hdlc.windowSizeTX = s.settings.hdlc.windowSizeTX
            elif s.command == Command.UA and s.settings.isServer:
                self.settings.hdlc.maxInfoTX = s.settings.hdlc.maxInfoTX
                self.settings.hdlc.maxInfoRX = s.settings.hdlc.maxInfoRX
                self.settings.hdlc.windowSizeRX = s.settings.hdlc.windowSizeRX
                self.settings.hdlc.windowSizeTX = s.settings.hdlc.windowSizeTX
            if s.template:
                reply = None
            p = GXDLMSXmlPdu(s.command, node, reply)
            if description:
                p.description = description
            if error:
                p.error = error
            if errorUrl:
                p.errorUrl = errorUrl
            if sleep:
                p.sleep = int(sleep)
            actions.append(p)
        return actions

    def pduToMessages(self, pdu):
        self.settings.command = pdu.command
        messages = []
        if pdu.command == Command.SNRM:
            messages.append(pdu.data)
        elif pdu.command == Command.UA:
            messages.append(pdu.data)
        elif pdu.command == Command.DISCONNECT_REQUEST:
            messages.append(GXDLMS.getHdlcFrame(self.settings, int(Command.DISCONNECT_REQUEST), GXByteBuffer(pdu.data)))
        else:
            reply = None
            if self.settings.interfaceType == InterfaceType.WRAPPER:
                if self.ciphering.security != Security.NONE:
                    p = GXDLMSLNParameters(self.settings, 0, pdu.command, 0x0, None, None, 0xff)
                    reply = GXByteBuffer(GXDLMS.cipher0(p, pdu.data))
                else:
                    reply = GXByteBuffer(pdu.data)
            else:
                if self.ciphering.security != Security.NONE:
                    p = GXDLMSLNParameters(self.settings, 0, pdu.command, 0x0, None, None, 0xff)
                    tmp = GXDLMS.cipher0(p, pdu.data)
                    reply = GXByteBuffer(len(tmp))
                    reply.set(_GXCommon.LLC_SEND_BYTES)
                    reply.set(tmp)
                else:
                    reply = GXByteBuffer(len(pdu.data))
                    reply.set(_GXCommon.LLC_SEND_BYTES)
                    reply.set(pdu.data)
            frame_ = 0
            while reply.position != len(reply):
                if self.settings.interfaceType == InterfaceType.WRAPPER:
                    messages.append(GXDLMS.getWrapperFrame(self.settings, pdu.command, reply))
                elif self.settings.interfaceType == InterfaceType.HDLC or self.settings.interfaceType == InterfaceType.HDLC_WITH_MODE_E:
                    if pdu.command == Command.AARQ:
                        frame_ = 0x10
                    elif pdu.command == Command.AARE:
                        frame_ = 0x30
                    elif pdu.command == Command.EVENT_NOTIFICATION:
                        frame_ = 0x13
                    messages.append(GXDLMS.getHdlcFrame(self.settings, frame_, reply))
                    if reply.position != len(reply):
                        frame_ = self.settings.getNextSend(False)
                elif self.settings.interfaceType == InterfaceType.PDU:
                    messages.append(reply.array())
                    break
                else:
                    raise ValueError("InterfaceType")
        return messages



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#

###Python 2 requires this
#pylint: disable=bad-option-value,old-style-class
class GXDLMSScheduleEntry:
    """Executed scripts."""
    #pylint: disable=too-few-public-methods,too-many-instance-attributes

    def __init__(self):
        """#Constructor."""
        # Schedule entry index.
        self.index = 0
        # Is Schedule entry enabled.
        self.enable = False
        # Logical name of the Script table object.
        self.logicalName = None
        # Script identifier of the script to be executed.
        self.scriptSelector = 0
        self.switchTime = None
        # Defines a period in minutes, in which an entry shall be processed
        # after power fail.
        self.validityWindow = 0
        # Days of the week on which the entry is valid.
        self.execWeekdays = None
        # Perform the link to the IC Special days table, day_id.
        self.execSpecDays = None
        # Date starting period in which the entry is valid.
        self.beginDate = None
        # Date starting period in which the entry is valid.
        self.endDate = None



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import ImageTransferStatus
from .GXDLMSImageActivateInfo import GXDLMSImageActivateInfo

# pylint: disable=too-many-instance-attributes
class GXDLMSImageTransfer(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSImageTransfer
    """

    def __init__(self, ln="0.0.44.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.IMAGE_TRANSFER, ln, sn)
        self.imageBlockSize = 200
        self.imageTransferredBlocksStatus = ""
        self.imageFirstNotTransferredBlockNumber = 0
        self.imageTransferEnabled = True
        self.imageActivateInfo = []
        self.imageTransferStatus = ImageTransferStatus.IMAGE_TRANSFER_NOT_INITIATED
        self.imageSize = 0
        self.imageData = None

    def getValues(self):
        return [self.logicalName,
                self.imageBlockSize,
                self.imageTransferredBlocksStatus,
                self.imageFirstNotTransferredBlockNumber,
                self.imageTransferEnabled,
                self.imageTransferStatus,
                self.imageActivateInfo]

    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  ImageBlockSize
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  ImageTransferredBlocksStatus
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  ImageFirstNotTransferredBlockNumber
        if all_ or not self.isRead(4):
            attributes.append(4)
        #  ImageTransferEnabled
        if all_ or not self.isRead(5):
            attributes.append(5)
        #  ImageTransferStatus
        if all_ or not self.isRead(6):
            attributes.append(6)
        #  ImageActivateInfo
        if all_ or not self.isRead(7):
            attributes.append(7)
        return attributes

    def getAttributeCount(self):
        return 7

    def getMethodCount(self):
        return 4

    def invoke(self, settings, e):
        #  Image transfer initiate
        if e.index == 1:
            self.imageFirstNotTransferredBlockNumber = 0
            self.imageTransferredBlocksStatus = ""
            value = e.parameters
            imageIdentifier = value[0]
            self.imageSize = value[1]
            self.imageTransferStatus = ImageTransferStatus.IMAGE_TRANSFER_INITIATED
            list_ = []
            list_.append(self.imageActivateInfo)
            item = None
            for it in self.imageActivateInfo:
                if it.identification == imageIdentifier:
                    item = it
                    break
            if item is None:
                item = GXDLMSImageActivateInfo()
                list_.append(item)
            item.size = self.imageSize
            item.identification = imageIdentifier
            self.imageActivateInfo = list_
            cnt = int((self.imageSize / self.imageBlockSize))
            if self.imageSize % self.imageBlockSize != 0:
                cnt += 1
            sb = ""
            pos = 0
            while pos < cnt:
                sb += '0'
                pos += 1
            self.imageTransferredBlocksStatus = sb
            return None
        if e.index == 2:
            #  Image block transfer
            value = e.parameters
            imageIndex = (value[0]).longValue()
            tmp = self.imageTransferredBlocksStatus.encode()
            tmp[int(imageIndex)] = '1'
            self.imageTransferredBlocksStatus = str(tmp)
            self.imageFirstNotTransferredBlockNumber = imageIndex + 1
            self.imageData.put(imageIndex, value[1])
            self.imageTransferStatus = ImageTransferStatus.IMAGE_TRANSFER_INITIATED
            return None
        if e.index == 3:
            #  Image verify.
            return None
        if e.index == 4:
            return None
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.UINT32
        elif index == 3:
            ret = DataType.BITSTRING
        elif index == 4:
            ret = DataType.UINT32
        elif index == 5:
            ret = DataType.BOOLEAN
        elif index == 6:
            ret = DataType.ENUM
        elif index == 7:
            ret = DataType.ARRAY
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.imageBlockSize
        elif e.index == 3:
            ret = self.imageTransferredBlocksStatus
        elif e.index == 4:
            ret = self.imageFirstNotTransferredBlockNumber
        elif e.index == 5:
            ret = self.imageTransferEnabled
        elif e.index == 6:
            ret = self.imageTransferStatus
        elif e.index == 7:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            #  Count
            data.setUInt8(len(self.imageActivateInfo))
            for it in self.imageActivateInfo:
                data.setUInt8(DataType.STRUCTURE)
                #  Item count.
                data.setUInt8(int(3))
                _GXCommon.setData(settings, data, DataType.UINT32, it.size)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, it.identification)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, it.signature)
            ret = data
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if e.value is None:
                self.imageBlockSize = 0
            else:
                self.imageBlockSize = e.value
        elif e.index == 3:
            self.imageTransferredBlocksStatus = str(e.value)
        elif e.index == 4:
            if e.value is None:
                self.imageFirstNotTransferredBlockNumber = 0
            else:
                self.imageFirstNotTransferredBlockNumber = e.value
        elif e.index == 5:
            if e.value is None:
                self.imageTransferEnabled = False
            else:
                self.imageTransferEnabled = e.value
        elif e.index == 6:
            #pylint: disable=bad-option-value,redefined-variable-type
            if e.value is None:
                self.imageTransferStatus = ImageTransferStatus.IMAGE_TRANSFER_NOT_INITIATED
            else:
                self.imageTransferStatus = e.value
        elif e.index == 7:
            self.imageActivateInfo = []
            if e.value:
                for it in e.value:
                    item = GXDLMSImageActivateInfo()
                    item.size = it[0]
                    item.identification = it[1]
                    item.signature = it[2]
                    self.imageActivateInfo.append(item)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def imageTransferInitiate(self, client, imageIdentifier, imageSize):
        if self.imageBlockSize == 0:
            raise ValueError("Invalid image block size.")
        if self.imageBlockSize > client.maxReceivePDUSize:
            raise ValueError("Image block size is bigger than max PDU size.")
        data = GXByteBuffer()
        data.setUInt8(DataType.STRUCTURE)
        data.setUInt8(2)
        _GXCommon.setData(None, data, DataType.OCTET_STRING, _GXCommon.getBytes(imageIdentifier))
        _GXCommon.setData(None, data, DataType.UINT32, imageSize)
        return client.method(self, 1, data, DataType.ARRAY)

    def imageBlockTransfer(self, client, imageBlockValue, imageBlockCount):
        cnt = int(len(imageBlockValue) / self.imageBlockSize)
        if (len(imageBlockValue) % self.imageBlockSize) != 0:
            cnt += 1
        if imageBlockCount:
            imageBlockCount[0] = cnt
        packets = []
        pos = 0
        while pos != cnt:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            _GXCommon.setData(None, data, DataType.UINT32, pos)
            tmp = None
            len_ = len(imageBlockValue) - ((pos + 1) * self.imageBlockSize)
            #  If last packet
            if len_ < 0:
                tmp = imageBlockValue[pos * self.imageBlockSize:]
            else:
                tmp = imageBlockValue[pos * self.imageBlockSize: pos * self.imageBlockSize + self.imageBlockSize]
            _GXCommon.setData(None, data, DataType.OCTET_STRING, tmp)
            packets.append(client.method(self, 2, data, DataType.ARRAY))
            pos += 1
        return packets

    def imageVerify(self, client):
        return client.method(self, 3, int(0), DataType.INT8)

    def imageActivate(self, client):
        return client.method(self, 4, int(0), DataType.INT8)

    def load(self, reader):
        self.imageBlockSize = reader.readElementContentAsInt("ImageBlockSize")
        self.imageTransferredBlocksStatus = reader.readElementContentAsString("ImageTransferredBlocksStatus")
        self.imageFirstNotTransferredBlockNumber = reader.readElementContentAsLong("ImageFirstNotTransferredBlockNumber")
        self.imageTransferEnabled = reader.readElementContentAsInt("ImageTransferEnabled") != 0
        self.imageTransferStatus = reader.readElementContentAsInt("ImageTransferStatus")
        self.imageActivateInfo = []
        if reader.isStartElement("ImageActivateInfo", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSImageActivateInfo()
                it.size = reader.readElementContentAsULong("Size")
                it.identification = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Identification"))
                it.signature = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Signature"))
                self.imageActivateInfo.append(it)
            reader.readEndElement("ImageActivateInfo")

    def save(self, writer):
        writer.writeElementString("ImageBlockSize", self.imageBlockSize)
        writer.writeElementString("ImageTransferredBlocksStatus", self.imageTransferredBlocksStatus)
        writer.writeElementString("ImageFirstNotTransferredBlockNumber", self.imageFirstNotTransferredBlockNumber)
        writer.writeElementString("ImageTransferEnabled", self.imageTransferEnabled)
        writer.writeElementString("ImageTransferStatus", int(self.imageTransferStatus))
        writer.writeStartElement("ImageActivateInfo")
        if self.imageActivateInfo:
            for it in self.imageActivateInfo:
                writer.writeStartElement("Item")
                writer.writeElementString("Size", it.size)
                writer.writeElementString("Identification", GXByteBuffer.hex(it.identification, False))
                writer.writeElementString("Signature", GXByteBuffer.hex(it.signature, False))
                writer.writeEndElement()
        writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import socket
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import AddressConfigMode
from .GXNeighborDiscoverySetup import GXNeighborDiscoverySetup


# pylint: disable=too-many-instance-attributes
class GXDLMSIp6Setup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSIp6Setup
    """

    def __init__(self, ln="0.0.25.7.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        # pylint: disable=super-with-arguments
        super(GXDLMSIp6Setup, self).__init__(ObjectType.IP6_SETUP, ln, sn)
        self.addressConfigMode = AddressConfigMode.AUTO
        self.dataLinkLayerReference = None
        self.unicastIPAddress = []
        self.multicastIPAddress = []
        self.gatewayIPAddress = []
        self.primaryDNSAddress = ""
        self.secondaryDNSAddress = ""
        self.trafficClass = 0
        self.neighborDiscoverySetup = []

    def getValues(self):
        return [
            self.logicalName,
            self.dataLinkLayerReference,
            self.addressConfigMode,
            self.unicastIPAddress,
            self.multicastIPAddress,
            self.gatewayIPAddress,
            self.primaryDNSAddress,
            self.secondaryDNSAddress,
            self.trafficClass,
            self.neighborDiscoverySetup,
        ]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  DataLinkLayerReference
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  AddressConfigMode
        if all_ or self.canRead(3):
            attributes.append(3)
        #  UnicastIPAddress
        if all_ or self.canRead(4):
            attributes.append(4)
        #  MulticastIPAddress
        if all_ or self.canRead(5):
            attributes.append(5)
        #  GatewayIPAddress
        if all_ or self.canRead(6):
            attributes.append(6)
        #  PrimaryDNSAddress
        if all_ or self.canRead(7):
            attributes.append(7)
        #  SecondaryDNSAddress
        if all_ or not self.isRead(8):
            attributes.append(8)
        #  TrafficClass
        if all_ or self.canRead(9):
            attributes.append(9)
        #  NeighborDiscoverySetup
        if all_ or self.canRead(10):
            attributes.append(10)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 10

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 2

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.OCTET_STRING
        elif index == 3:
            ret = DataType.ENUM
        elif index == 4:
            ret = DataType.ARRAY
        elif index == 5:
            ret = DataType.ARRAY
        elif index == 6:
            ret = DataType.ARRAY
        elif index == 7:
            ret = DataType.OCTET_STRING
        elif index == 8:
            ret = DataType.OCTET_STRING
        elif index == 9:
            ret = DataType.UINT8
        elif index == 10:
            ret = DataType.ARRAY
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = _GXCommon.logicalNameToBytes(self.dataLinkLayerReference)
        elif e.index == 3:
            ret = self.addressConfigMode
        elif e.index == 4:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if not self.unicastIPAddress:
                _GXCommon.setObjectCount(0, data)
            else:
                _GXCommon.setObjectCount(len(self.unicastIPAddress), data)
                for it in self.unicastIPAddress:
                    _GXCommon.setData(
                        settings,
                        data,
                        DataType.OCTET_STRING,
                        socket.inet_pton(socket.AF_INET6, it),
                    )
            ret = data
        elif e.index == 5:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if not self.multicastIPAddress:
                _GXCommon.setObjectCount(0, data)
            else:
                _GXCommon.setObjectCount(len(self.multicastIPAddress), data)
                for it in self.multicastIPAddress:
                    _GXCommon.setData(
                        settings,
                        data,
                        DataType.OCTET_STRING,
                        socket.inet_pton(socket.AF_INET6, it),
                    )
            ret = data
        elif e.index == 6:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if not self.gatewayIPAddress:
                _GXCommon.setObjectCount(0, data)
            else:
                _GXCommon.setObjectCount(len(self.gatewayIPAddress), data)
                for it in self.gatewayIPAddress:
                    _GXCommon.setData(
                        settings,
                        data,
                        DataType.OCTET_STRING,
                        socket.inet_pton(socket.AF_INET6, it),
                    )
            ret = data
        elif e.index == 7:
            if not self.primaryDNSAddress:
                ret = None
            else:
                ret = socket.inet_pton(socket.AF_INET6, self.primaryDNSAddress)
        elif e.index == 8:
            if not self.secondaryDNSAddress:
                ret = None
            else:
                ret = socket.inet_pton(socket.AF_INET6, self.secondaryDNSAddress)
        elif e.index == 9:
            ret = self.trafficClass
        elif e.index == 10:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if self.neighborDiscoverySetup is None:
                #  Object count is zero.
                data.setUInt8(0)
            else:
                _GXCommon.setObjectCount(len(self.neighborDiscoverySetup), data)
                for it in self.neighborDiscoverySetup:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(3)
                    _GXCommon.setData(settings, data, DataType.UINT8, it.maxRetry)
                    _GXCommon.setData(settings, data, DataType.UINT16, it.retryWaitTime)
                    _GXCommon.setData(settings, data, DataType.UINT32, it.sendPeriod)
            ret = data
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if isinstance(e.value, str):
                self.dataLinkLayerReference = str(e.value)
            else:
                self.dataLinkLayerReference = _GXCommon.toLogicalName(e.value)
        elif e.index == 3:
            self.addressConfigMode = AddressConfigMode(e.value)
        elif e.index == 4:
            self.unicastIPAddress = []
            if e.value:
                # This fails in Python 2.7.4. Update to 2.7.6
                # https://bugs.python.org/issue10212
                for it in e.value:
                    self.unicastIPAddress.append(socket.inet_ntop(socket.AF_INET6, it))
        elif e.index == 5:
            self.multicastIPAddress = []
            if e.value:
                # This fails in Python 2.7.4. Update to 2.7.6
                # https://bugs.python.org/issue10212
                for it in e.value:
                    self.multicastIPAddress.append(
                        socket.inet_ntop(socket.AF_INET6, it)
                    )
        elif e.index == 6:
            self.gatewayIPAddress = []
            if e.value:
                # This fails in Python 2.7.4. Update to 2.7.6
                # https://bugs.python.org/issue10212
                for it in e.value:
                    self.gatewayIPAddress.append(socket.inet_ntop(socket.AF_INET6, it))
        elif e.index == 7:
            if not e.value:
                self.primaryDNSAddress = None
            else:
                self.primaryDNSAddress = socket.inet_ntop(socket.AF_INET6, e.value)
        elif e.index == 8:
            if not e.value:
                self.secondaryDNSAddress = None
            else:
                # This fails in Python 2.7.4. Update to 2.7.6
                # https://bugs.python.org/issue10212
                self.secondaryDNSAddress = socket.inet_ntop(socket.AF_INET6, e.value)
        elif e.index == 9:
            self.trafficClass = e.value
        elif e.index == 10:
            self.neighborDiscoverySetup = []
            if e.value:
                for it in e.value:
                    v = GXNeighborDiscoverySetup()
                    v.maxRetry = it[0]
                    v.retryWaitTime = it[1]
                    v.sendPeriod = it[2]
                    self.neighborDiscoverySetup.append(v)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @classmethod
    def loadIPAddress(cls, reader, name):
        list_ = []
        if reader.isStartElement(name, True):
            while reader.isStartElement("Value", False):
                list_.append(reader.readElementContentAsString("Value"))
            reader.readEndElement(name)
        return list_

    @classmethod
    def loadNeighborDiscoverySetup(cls, reader, name):
        list_ = []
        if reader.isStartElement(name, True):
            while reader.isStartElement("Item", True):
                it = GXNeighborDiscoverySetup()
                list_.append(it)
                it.maxRetry = reader.readElementContentAsInt("MaxRetry")
                it.retryWaitTime = reader.readElementContentAsInt("RetryWaitTime")
                it.sendPeriod = reader.readElementContentAsInt("SendPeriod")
            reader.readEndElement(name)
        return list_

    def load(self, reader):
        self.dataLinkLayerReference = reader.readElementContentAsString(
            "DataLinkLayerReference"
        )
        self.addressConfigMode = reader.readElementContentAsInt("AddressConfigMode")
        self.unicastIPAddress = self.loadIPAddress(reader, "UnicastIPAddress")
        self.multicastIPAddress = self.loadIPAddress(reader, "MulticastIPAddress")
        self.gatewayIPAddress = self.loadIPAddress(reader, "GatewayIPAddress")
        self.primaryDNSAddress = reader.readElementContentAsString("PrimaryDNSAddress")
        self.secondaryDNSAddress = reader.readElementContentAsString(
            "SecondaryDNSAddress"
        )
        self.trafficClass = reader.readElementContentAsInt("TrafficClass")
        self.neighborDiscoverySetup = self.loadNeighborDiscoverySetup(
            reader, "NeighborDiscoverySetup"
        )

    @classmethod
    def saveIPAddress(cls, writer, list_, name):
        writer.writeStartElement(name)
        if list_:
            for it in list_:
                writer.writeElementString("Value", it)
        writer.writeEndElement()

    @classmethod
    def saveNeighborDiscoverySetup(cls, writer, list_, name):
        writer.writeStartElement(name)
        if list_:
            for it in list_:
                writer.writeStartElement("Item")
                writer.writeElementString("MaxRetry", it.maxRetry)
                writer.writeElementString("RetryWaitTime", it.retryWaitTime)
                writer.writeElementString("SendPeriod", it.sendPeriod)
                writer.writeEndElement()
        writer.writeEndElement()

    def save(self, writer):
        writer.writeElementString("DataLinkLayerReference", self.dataLinkLayerReference)
        writer.writeElementString("AddressConfigMode", int(self.addressConfigMode))
        self.saveIPAddress(writer, self.unicastIPAddress, "UnicastIPAddress")
        self.saveIPAddress(writer, self.multicastIPAddress, "MulticastIPAddress")
        self.saveIPAddress(writer, self.gatewayIPAddress, "GatewayIPAddress")
        writer.writeElementString("PrimaryDNSAddress", self.primaryDNSAddress)
        writer.writeElementString("SecondaryDNSAddress", self.secondaryDNSAddress)
        writer.writeElementString("TrafficClass", self.trafficClass)
        self.saveNeighborDiscoverySetup(
            writer, self.neighborDiscoverySetup, "NeighborDiscoverySetup"
        )



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import DataType, ObjectType, Authentication
from .enums import AssociationStatus
from .GXDLMSObjectCollection import GXDLMSObjectCollection
from .GXApplicationContextName import GXApplicationContextName
from .GXxDLMSContextType import GXxDLMSContextType
from .GXAuthenticationMechanismName import GXAuthenticationMechanismName
from ..ValueEventArgs import ValueEventArgs
from ..GXSecure import GXSecure
from ..enums.Conformance import Conformance
from ..enums.AccessMode import AccessMode
from ..enums.MethodAccessMode import MethodAccessMode
from ..enums.AccessMode3 import AccessMode3
from ..enums.MethodAccessMode3 import MethodAccessMode3
from ..GXBitString import GXBitString


# pylint: disable=too-many-instance-attributes
class GXDLMSAssociationLogicalName(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSAssociationLogicalName
    """

    def __init__(self, ln="0.0.40.0.0.255"):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.ASSOCIATION_LOGICAL_NAME, ln)
        self.objectList = GXDLMSObjectCollection(self)
        self.applicationContextName = GXApplicationContextName()
        self.xDLMSContextInfo = GXxDLMSContextType()
        self.authenticationMechanismName = GXAuthenticationMechanismName()
        self.userList = []
        self.version = 2
        self.associationStatus = AssociationStatus.NON_ASSOCIATED
        self.clientSAP = None
        self.serverSAP = None
        self.secret = None
        self.securitySetupReference = None
        self.userList = []
        self.currentUser = None

    #
    # Updates secret.
    #
    # @param client
    # DLMS client.
    # Action bytes.
    #
    def updateSecret(self, client):
        if self.authenticationMechanismName.mechanismId == Authentication.NONE:
            raise ValueError("Invalid authentication level in MechanismId.")
        if self.authenticationMechanismName.mechanismId == Authentication.HIGH_GMAC:
            raise ValueError("HighGMAC secret is updated using Security setup.")
        if self.authenticationMechanismName.mechanismId == Authentication.LOW:
            return client.write(self, 7)
        #  Action is used to update High authentication password.
        return client.method(self, 2, self.secret, DataType.OCTET_STRING)

    #
    # Add user to user list.
    #
    # @param client
    # DLMS client.
    # @param id
    # User ID.
    # @param name
    # User name.
    # Action bytes.
    #
    def addUser(self, client, id_, name):
        data = GXByteBuffer()
        data.setUInt8(DataType.STRUCTURE)
        #  Add structure size.
        data.setUInt8(2)
        _GXCommon.setData(None, data, DataType.UINT8, id_)
        _GXCommon.setData(None, data, DataType.STRING, name)
        return client.method(self, 5, data.array(), DataType.STRUCTURE)

    #
    # Remove user from user list.
    #
    # @param client
    # DLMS client.
    # @param id
    # User ID.
    # @param name
    # User name.
    # Action bytes.
    #
    def removeUser(self, client, id_, name):
        data = GXByteBuffer()
        data.setUInt8(DataType.STRUCTURE)
        #  Add structure size.
        data.setUInt8(2)
        _GXCommon.setData(None, data, DataType.UINT8, id_)
        _GXCommon.setData(None, data, DataType.STRING, name)
        return client.method(self, 6, data.array(), DataType.STRUCTURE)

    def getValues(self):
        return [
            self.logicalName,
            self.objectList,
            [self.clientSAP, self.serverSAP],
            self.applicationContextName,
            self.xDLMSContextInfo,
            self.authenticationMechanismName,
            self.secret,
            self.associationStatus,
            self.securitySetupReference,
            self.userList,
            self.currentUser,
        ]

    def invoke(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        #  Check reply_to_HLS_authentication
        if e.index == 1:
            serverChallenge = None
            clientChallenge = None
            ic = 0
            readSecret = []
            accept = True
            if settings.authentication == Authentication.HIGH_ECDSA:
                raise ValueError("ECDSA is not implemented.")
            if settings.authentication == Authentication.HIGH_GMAC:
                readSecret = settings.sourceSystemTitle
                bb = GXByteBuffer(int(e.parameters))
                bb.getUInt8()
                ic = bb.getUInt32()
            else:
                readSecret = self.secret
            serverChallenge = GXSecure.secure(
                settings, settings.cipher, ic, settings.stoCChallenge, readSecret
            )
            clientChallenge = int(e.parameters)
            accept = serverChallenge == clientChallenge
            if accept:
                if settings.authentication in (
                    Authentication.HIGH_GMAC,
                    Authentication.HIGH_ECDSA,
                ):
                    readSecret = settings.cipher.getSystemTitle()
                    ic = settings.cipher.invocationCounter
                else:
                    readSecret = self.secret
                tmp = GXSecure.secure(
                    settings,
                    settings.cipher,
                    ic,
                    settings.getCtoSChallenge(),
                    readSecret,
                )
                self.associationStatus = AssociationStatus.ASSOCIATED
                return tmp
            self.associationStatus = AssociationStatus.NON_ASSOCIATED
        elif e.index == 2:
            if not e.parameters:
                e.error = ErrorCode.READ_WRITE_DENIED
            else:
                self.secret = e.parameters
        elif e.index == 5:
            if not e.parameters:
                e.error = ErrorCode.READ_WRITE_DENIED
            else:
                self.userList.append((e.parameters[0], str(e.parameters[1])))
        elif e.index == 6:
            tmp = e.parameters
            if not tmp:
                e.error = ErrorCode.READ_WRITE_DENIED
            else:
                id_ = tmp[0]
                name = str(tmp[1])
                for k, v in self.userList:
                    if k == id_ and v == name:
                        self.userList.remove(k)
                        break
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return None

    def getAttributeIndexToRead(self, all_):
        attributes = []
        if all_ or not self.logicalName:
            attributes.append(1)
        if all_ or not self.isRead(2):
            attributes.append(2)
        if all_ or not self.isRead(3):
            attributes.append(3)
        if all_ or not self.isRead(4):
            attributes.append(4)
        if all_ or not self.isRead(5):
            attributes.append(5)
        if all_ or not self.isRead(6):
            attributes.append(6)
        if all_ or not self.isRead(7):
            attributes.append(7)
        if all_ or not self.isRead(8):
            attributes.append(8)
        if self.version > 0 and (all_ or not self.isRead(9)):
            attributes.append(9)
        if self.version > 1:
            if all_ or not self.isRead(10):
                attributes.append(10)
            if all_ or not self.isRead(11):
                attributes.append(11)
        return attributes

    def getAttributeCount(self):
        if self.version > 1:
            return 11
        if self.version > 0:
            return 9
        return 8

    def getMethodCount(self):
        if self.version > 1:
            return 6
        return 4

    def getObjects(self, settings, e):
        data = GXByteBuffer()
        if settings.index == 0:
            settings.setCount(len(self.objectList))
            data.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.objectList), data)
        pos = 0
        for it in self.objectList:
            pos += 1
            if not pos <= settings.index:
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(4)
                _GXCommon.setData(settings, data, DataType.UINT16, it.objectType)
                _GXCommon.setData(settings, data, DataType.UINT8, it.version)
                _GXCommon.setData(
                    settings,
                    data,
                    DataType.OCTET_STRING,
                    _GXCommon.logicalNameToBytes(it.logicalName),
                )
                self.__getAccessRights(settings, it, e.server, data)
                settings.index = settings.index + 1
                if settings.isServer:
                    if not e.isSkipMaxPduSize and len(data) >= settings.maxPduSize:
                        break
        return data

    def __getAccessRights(self, settings, item, server, data):
        data.setUInt8(DataType.STRUCTURE)
        data.setUInt8(2)
        if server is None:
            data.setUInt8(DataType.ARRAY)
            data.setUInt8(0)
            data.setUInt8(DataType.ARRAY)
            data.setUInt8(0)
        else:
            data.setUInt8(DataType.ARRAY)
            cnt = item.getAttributeCount()
            data.setUInt8(cnt)
            e = ValueEventArgs(server, item, 0, 0, None)
            pos = 0
            while pos != cnt:
                e.index = pos + 1
                m = server.onGetAttributeAccess(e)
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(3)
                _GXCommon.setData(settings, data, DataType.INT8, pos + 1)
                _GXCommon.setData(settings, data, DataType.ENUM, m)
                _GXCommon.setData(settings, data, DataType.NONE, None)
                pos += 1
            data.setUInt8(DataType.ARRAY)
            cnt = item.getMethodCount()
            data.setUInt8(cnt)
            pos = 0
            while pos != cnt:
                e.index = pos + 1
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(2)
                _GXCommon.setData(settings, data, DataType.INT8, pos + 1)
                m = server.onGetMethodAccess(e)
                if self.version == 0:
                    _GXCommon.setData(settings, data, DataType.BOOLEAN, m != 0)
                else:
                    _GXCommon.setData(settings, data, DataType.ENUM, m)
                pos += 1

    def updateAccessRights(self, obj, buff):
        if buff:
            for attributeAccess in buff[0]:
                id_ = attributeAccess[0]
                tmp = attributeAccess[1]
                if self.version < 3:
                    obj.setAccess(id_, AccessMode(tmp))
                else:
                    obj.setAccess(id_, AccessMode3(tmp))

            for methodAccess in buff[1]:
                id_ = methodAccess[0]
                tmp = 0
                if isinstance(methodAccess[1], bool):
                    if methodAccess[1]:
                        tmp = 1
                    else:
                        tmp = 0
                else:
                    tmp = methodAccess[1]
                if self.version < 3:
                    obj.setMethodAccess(id_, MethodAccessMode(tmp))
                else:
                    obj.setMethodAccess3(id_, MethodAccessMode3(tmp))

    def getUserList(self, settings):
        data = GXByteBuffer()
        if settings.index == 0:
            settings.setCount(len(self.userList))
            data.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.userList), data)
        pos = 0
        for k, v in self.userList:
            pos += 1
            if not pos <= settings.index:
                settings.index = settings.index + 1
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(2)
                _GXCommon.setData(settings, data, DataType.UINT8, k)
                _GXCommon.setData(settings, data, DataType.STRING, v)
        return data

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ARRAY
        elif index == 3:
            ret = DataType.STRUCTURE
        elif index == 4:
            ret = DataType.STRUCTURE
        elif index == 5:
            ret = DataType.STRUCTURE
        elif index == 6:
            ret = DataType.STRUCTURE
        elif index == 7:
            ret = DataType.OCTET_STRING
        elif index == 8:
            ret = DataType.ENUM
        elif self.version > 0 and index == 9:
            ret = DataType.OCTET_STRING
        elif self.version > 1:
            if index == 10:
                ret = DataType.ARRAY
            if index == 11:
                ret = DataType.STRUCTURE
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    def getValue(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.getObjects(settings, e)
        elif e.index == 3:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            data.setUInt8(DataType.INT8)
            data.setUInt8(self.clientSAP)
            data.setUInt8(DataType.UINT16)
            data.setUInt16(self.serverSAP)
            ret = data
        elif e.index == 4:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(0x7)
            _GXCommon.setData(
                settings, data, DataType.UINT8, self.applicationContextName.jointIsoCtt
            )
            _GXCommon.setData(
                settings, data, DataType.UINT8, self.applicationContextName.country
            )
            _GXCommon.setData(
                settings, data, DataType.UINT16, self.applicationContextName.countryName
            )
            _GXCommon.setData(
                settings,
                data,
                DataType.UINT8,
                self.applicationContextName.identifiedOrganization,
            )
            _GXCommon.setData(
                settings, data, DataType.UINT8, self.applicationContextName.dlmsUA
            )
            _GXCommon.setData(
                settings,
                data,
                DataType.UINT8,
                self.applicationContextName.applicationContext,
            )
            _GXCommon.setData(
                settings, data, DataType.UINT8, self.applicationContextName.contextId
            )
            ret = data
        elif e.index == 5:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(6)
            _GXCommon.setData(
                settings,
                data,
                DataType.BITSTRING,
                GXBitString.toBitString(self.xDLMSContextInfo.conformance, 24),
            )
            _GXCommon.setData(
                settings, data, DataType.UINT16, self.xDLMSContextInfo.maxReceivePduSize
            )
            _GXCommon.setData(
                settings, data, DataType.UINT16, self.xDLMSContextInfo.maxSendPduSize
            )
            _GXCommon.setData(
                settings, data, DataType.UINT8, self.xDLMSContextInfo.dlmsVersionNumber
            )
            _GXCommon.setData(
                settings, data, DataType.INT8, self.xDLMSContextInfo.qualityOfService
            )
            _GXCommon.setData(
                settings,
                data,
                DataType.OCTET_STRING,
                self.xDLMSContextInfo.cypheringInfo,
            )
            ret = data
        elif e.index == 6:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(0x7)
            _GXCommon.setData(
                settings,
                data,
                DataType.UINT8,
                self.authenticationMechanismName.jointIsoCtt,
            )
            _GXCommon.setData(
                settings, data, DataType.UINT8, self.authenticationMechanismName.country
            )
            _GXCommon.setData(
                settings,
                data,
                DataType.UINT16,
                self.authenticationMechanismName.countryName,
            )
            _GXCommon.setData(
                settings,
                data,
                DataType.UINT8,
                self.authenticationMechanismName.identifiedOrganization,
            )
            _GXCommon.setData(
                settings, data, DataType.UINT8, self.authenticationMechanismName.dlmsUA
            )
            _GXCommon.setData(
                settings,
                data,
                DataType.UINT8,
                self.authenticationMechanismName.authenticationMechanismName,
            )
            _GXCommon.setData(
                settings,
                data,
                DataType.UINT8,
                self.authenticationMechanismName.mechanismId,
            )
            ret = data
        elif e.index == 7:
            ret = self.secret
        elif e.index == 8:
            ret = self.associationStatus
        elif e.index == 9:
            ret = _GXCommon.logicalNameToBytes(self.securitySetupReference)
        elif e.index == 10:
            ret = self.getUserList(settings)
        elif e.index == 11:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            if self.currentUser is None:
                _GXCommon.setData(settings, data, DataType.UINT8, 0)
                _GXCommon.setData(settings, data, DataType.STRING, None)
            else:
                _GXCommon.setData(settings, data, DataType.UINT8, self.currentUser[0])
                _GXCommon.setData(settings, data, DataType.STRING, self.currentUser[1])
            ret = data
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    def updateObjectList(self, settings, target, value):
        # pylint: disable=import-outside-toplevel,unidiomatic-typecheck
        from .._GXObjectFactory import _GXObjectFactory

        target.clear()
        if value:
            for item in value:
                type_ = item[0]
                version = item[1]
                ln = _GXCommon.toLogicalName(item[2])
                obj = settings.objects.findByLN(type_, ln)
                if obj is None:
                    obj = _GXObjectFactory.createObject(type_)
                    obj.logicalName = ln
                    obj.version = version
                if type(obj) != GXDLMSObject:
                    self.updateAccessRights(obj, item[3])
                    target.append(obj)

    def updateApplicationContextName(self, value):
        if isinstance(value, bytearray):
            buff = GXByteBuffer(value)
            if buff.getUInt8(0) == 0x60:
                self.applicationContextName.jointIsoCtt = 0
                self.applicationContextName.country = 0
                self.applicationContextName.countryName = 0
                buff.position = buff.position + 3
                self.applicationContextName.identifiedOrganization = buff.getUInt8()
                self.applicationContextName.dlmsUA = buff.getUInt8()
                self.applicationContextName.applicationContext = buff.getUInt8()
                self.applicationContextName.contextId = buff.getUInt8()
            else:
                if buff.getUInt8() != 2 and buff.getUInt8() != 7:
                    raise ValueError()
                if buff.getUInt8() != 0x11:
                    raise ValueError()
                self.applicationContextName.jointIsoCtt = buff.getUInt8()
                if buff.getUInt8() != 0x11:
                    raise ValueError()
                self.applicationContextName.country = buff.getUInt8()
                if buff.getUInt8() != 0x12:
                    raise ValueError()
                self.applicationContextName.countryName = buff.getUInt16()
                if buff.getUInt8() != 0x11:
                    raise ValueError()
                self.applicationContextName.identifiedOrganization = buff.getUInt8()
                if buff.getUInt8() != 0x11:
                    raise ValueError()
                self.applicationContextName.dlmsUA = buff.getUInt8()
                if buff.getUInt8() != 0x11:
                    raise ValueError()
                self.applicationContextName.applicationContext = buff.getUInt8()
                if buff.getUInt8() != 0x11:
                    raise ValueError()
                self.applicationContextName.contextId = buff.getUInt8()
        else:
            if value:
                self.applicationContextName.jointIsoCtt = value[0]
                self.applicationContextName.country = value[1]
                self.applicationContextName.countryName = value[2]
                self.applicationContextName.identifiedOrganization = value[3]
                self.applicationContextName.dlmsUA = value[4]
                self.applicationContextName.applicationContext = value[5]
                self.applicationContextName.contextId = value[6]

    def updateAuthenticationMechanismName(self, value):
        if value:
            if isinstance(value, bytearray):
                buff = GXByteBuffer(value)
                if buff.getUInt8(0) == 0x60:
                    self.authenticationMechanismName.jointIsoCtt = 0
                    self.authenticationMechanismName.country = 0
                    self.authenticationMechanismName.countryName = 0
                    buff.position = buff.position + 3
                    self.authenticationMechanismName.identifiedOrganization = (
                        buff.getUInt8()
                    )
                    self.authenticationMechanismName.dlmsUA = buff.getUInt8()
                    self.authenticationMechanismName.authenticationMechanismName = (
                        buff.getUInt8()
                    )
                    self.authenticationMechanismName.mechanismId = buff.getUInt8()
                else:
                    if buff.getUInt8() != 2 and buff.getUInt8() != 7:
                        raise ValueError()
                    if buff.getUInt8() != 0x11:
                        raise ValueError()
                    self.authenticationMechanismName.jointIsoCtt = buff.getUInt8()
                    if buff.getUInt8() != 0x11:
                        raise ValueError()
                    self.authenticationMechanismName.country = buff.getUInt8()
                    if buff.getUInt8() != 0x12:
                        raise ValueError()
                    self.authenticationMechanismName.countryName = buff.getUInt16()
                    if buff.getUInt8() != 0x11:
                        raise ValueError()
                    self.authenticationMechanismName.identifiedOrganization = (
                        buff.getUInt8()
                    )
                    if buff.getUInt8() != 0x11:
                        raise ValueError()
                    self.authenticationMechanismName.dlmsUA = buff.getUInt8()
                    if buff.getUInt8() != 0x11:
                        raise ValueError()
                    self.authenticationMechanismName.authenticationMechanismName = (
                        buff.getUInt8()
                    )
                    if buff.getUInt8() != 0x11:
                        raise ValueError()
                    self.authenticationMechanismName.mechanismId = buff.getUInt8()
            else:
                if value:
                    self.authenticationMechanismName.jointIsoCtt = value[0]
                    self.authenticationMechanismName.country = value[1]
                    self.authenticationMechanismName.countryName = value[2]
                    self.authenticationMechanismName.identifiedOrganization = value[3]
                    self.authenticationMechanismName.dlmsUA = value[4]
                    self.authenticationMechanismName.authenticationMechanismName = (
                        value[5]
                    )
                    self.authenticationMechanismName.mechanismId = value[6]

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.updateObjectList(settings, self.objectList, e.value)
        elif e.index == 3:
            if e.value:
                self.clientSAP = e.value[0]
                self.serverSAP = e.value[1]
        elif e.index == 4:
            self.updateApplicationContextName(e.value)
        elif e.index == 5:
            if e.value:
                self.xDLMSContextInfo.conformance = Conformance(e.value[0].toInteger())
                self.xDLMSContextInfo.maxReceivePduSize = e.value[1]
                self.xDLMSContextInfo.maxSendPduSize = e.value[2]
                self.xDLMSContextInfo.dlmsVersionNumber = e.value[3]
                self.xDLMSContextInfo.qualityOfService = e.value[4]
                self.xDLMSContextInfo.cypheringInfo = e.value[5]
        elif e.index == 6:
            self.updateAuthenticationMechanismName(e.value)
        elif e.index == 7:
            self.secret = e.value
        elif e.index == 8:
            # pylint: disable=bad-option-value,redefined-variable-type
            if e.value is None:
                self.associationStatus = AssociationStatus.NON_ASSOCIATED
            else:
                self.associationStatus = e.value
        elif e.index == 9:
            self.securitySetupReference = _GXCommon.toLogicalName(e.value)
        elif e.index == 10:
            self.userList = []
            if e.value:
                for tmp in e.value:
                    item = tmp
                    self.userList.append((int(item[0]), str(item[1])))
        elif e.index == 11:
            if e.value:
                tmp = e.value
                self.currentUser = (tmp[0], str(tmp[1]))
            else:
                self.currentUser = None
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.clientSAP = reader.readElementContentAsInt("ClientSAP")
        self.serverSAP = reader.readElementContentAsInt("ServerSAP")
        if reader.isStartElement("ApplicationContextName", True):
            self.applicationContextName.jointIsoCtt = reader.readElementContentAsInt(
                "JointIsoCtt"
            )
            self.applicationContextName.country = reader.readElementContentAsInt(
                "Country"
            )
            self.applicationContextName.countryName = reader.readElementContentAsInt(
                "CountryName"
            )
            self.applicationContextName.identifiedOrganization = (
                reader.readElementContentAsInt("IdentifiedOrganization")
            )
            self.applicationContextName.dlmsUA = reader.readElementContentAsInt(
                "DlmsUA"
            )
            self.applicationContextName.applicationContext = (
                reader.readElementContentAsInt("ApplicationContext")
            )
            self.applicationContextName.contextId = reader.readElementContentAsInt(
                "ContextId"
            )
            reader.readEndElement("ApplicationContextName")
        if reader.isStartElement("XDLMSContextInfo", True):
            self.xDLMSContextInfo.conformance = Conformance(
                reader.readElementContentAsInt("Conformance")
            )
            self.xDLMSContextInfo.maxReceivePduSize = reader.readElementContentAsInt(
                "MaxReceivePduSize"
            )
            self.xDLMSContextInfo.maxSendPduSize = reader.readElementContentAsInt(
                "MaxSendPduSize"
            )
            self.xDLMSContextInfo.dlmsVersionNumber = reader.readElementContentAsInt(
                "DlmsVersionNumber"
            )
            self.xDLMSContextInfo.qualityOfService = reader.readElementContentAsInt(
                "QualityOfService"
            )
            self.xDLMSContextInfo.cypheringInfo = GXByteBuffer.hexToBytes(
                reader.readElementContentAsString("CypheringInfo")
            )
            reader.readEndElement("XDLMSContextInfo")
        if reader.isStartElement("AuthenticationMechanismName", True):
            self.authenticationMechanismName.JointIsoCtt = (
                reader.readElementContentAsInt("JointIsoCtt")
            )
            self.authenticationMechanismName.country = reader.readElementContentAsInt(
                "Country"
            )
            self.authenticationMechanismName.countryName = (
                reader.readElementContentAsInt("CountryName")
            )
            self.authenticationMechanismName.identifiedOrganization = (
                reader.readElementContentAsInt("IdentifiedOrganization")
            )
            self.authenticationMechanismName.dlmsUA = reader.readElementContentAsInt(
                "DlmsUA"
            )
            self.authenticationMechanismName.authenticationMechanismName = (
                reader.readElementContentAsInt("AuthenticationMechanismName")
            )
            self.authenticationMechanismName.mechanismId = (
                reader.readElementContentAsInt("MechanismId")
            )
            reader.readEndElement("AuthenticationMechanismName")
        str_ = reader.readElementContentAsString("Secret")
        if str_ is None:
            self.secret = None
        else:
            self.secret = GXByteBuffer.hexToBytes(str_)
        self.associationStatus = reader.readElementContentAsInt("AssociationStatus")
        self.securitySetupReference = reader.readElementContentAsString(
            "SecuritySetupReference"
        )

    def save(self, writer):
        writer.writeElementString("ClientSAP", self.clientSAP)
        writer.writeElementString("ServerSAP", self.serverSAP)
        if self.applicationContextName:
            writer.writeStartElement("ApplicationContextName")
            writer.writeElementString(
                "JointIsoCtt", self.applicationContextName.jointIsoCtt
            )
            writer.writeElementString("Country", self.applicationContextName.country)
            writer.writeElementString(
                "CountryName", self.applicationContextName.countryName
            )
            writer.writeElementString(
                "IdentifiedOrganization",
                self.applicationContextName.identifiedOrganization,
            )
            writer.writeElementString("DlmsUA", self.applicationContextName.dlmsUA)
            writer.writeElementString(
                "ApplicationContext", self.applicationContextName.applicationContext
            )
            writer.writeElementString(
                "ContextId", int(self.applicationContextName.contextId)
            )
            writer.writeEndElement()
        if self.xDLMSContextInfo:
            writer.writeStartElement("XDLMSContextInfo")
            writer.writeElementString(
                "Conformance", int(self.xDLMSContextInfo.conformance)
            )
            writer.writeElementString(
                "MaxReceivePduSize", self.xDLMSContextInfo.maxReceivePduSize
            )
            writer.writeElementString(
                "MaxSendPduSize", self.xDLMSContextInfo.maxSendPduSize
            )
            writer.writeElementString(
                "DlmsVersionNumber", self.xDLMSContextInfo.dlmsVersionNumber
            )
            writer.writeElementString(
                "QualityOfService", self.xDLMSContextInfo.qualityOfService
            )
            writer.writeElementString(
                "CypheringInfo", GXByteBuffer.hex(self.xDLMSContextInfo.cypheringInfo)
            )
            writer.writeEndElement()
        if self.authenticationMechanismName:
            writer.writeStartElement("AuthenticationMechanismName")
            writer.writeElementString(
                "JointIsoCtt", self.authenticationMechanismName.jointIsoCtt
            )
            writer.writeElementString(
                "Country", self.authenticationMechanismName.country
            )
            writer.writeElementString(
                "CountryName", self.authenticationMechanismName.countryName
            )
            writer.writeElementString(
                "IdentifiedOrganization",
                self.authenticationMechanismName.identifiedOrganization,
            )
            writer.writeElementString("DlmsUA", self.authenticationMechanismName.dlmsUA)
            writer.writeElementString(
                "AuthenticationMechanismName",
                self.authenticationMechanismName.authenticationMechanismName,
            )
            writer.writeElementString(
                "MechanismId", int(self.authenticationMechanismName.mechanismId)
            )
            writer.writeEndElement()
        writer.writeElementString("Secret", GXByteBuffer.hex(self.secret))
        writer.writeElementString("AssociationStatus", int(self.associationStatus))
        writer.writeElementString("SecuritySetupReference", self.securitySetupReference)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums import Security
from .GXICipher import GXICipher
from .objects.enums import SecuritySuite
from .GXDLMSChippering import GXDLMSChippering
from .AesGcmParameter import AesGcmParameter
from .GXByteBuffer import GXByteBuffer


# pylint: disable=too-many-instance-attributes, too-many-function-args, too-many-public-methods
class GXCiphering(GXICipher):
    """
    Gurux DLMS/COSEM Transport security (Ciphering) settings.
    """

    def __init__(self, title):
        """
        # Constructor. Default values are from the Green Book.
        title: Used system title.
        """
        self.publicKeys = []
        self.certificates = []
        self.security = Security.NONE
        # System title.
        self.systemTitle = title
        # Block cipher key.
        self.blockCipherKey = bytearray(
            (
                0x00,
                0x01,
                0x02,
                0x03,
                0x04,
                0x05,
                0x06,
                0x07,
                0x08,
                0x09,
                0x0A,
                0x0B,
                0x0C,
                0x0D,
                0x0E,
                0x0F,
            )
        )
        self.authenticationKey = bytearray(
            (
                0xD0,
                0xD1,
                0xD2,
                0xD3,
                0xD4,
                0xD5,
                0xD6,
                0xD7,
                0xD8,
                0xD9,
                0xDA,
                0xDB,
                0xDC,
                0xDD,
                0xDE,
                0xDF,
            )
        )
        # Dedicated key.
        self.dedicatedKey = None
        # Certificates.
        self.certificates = None
        # Ephemeral key pair.
        self.ephemeralKeyPair = None
        # recipient system title.
        self.recipientSystemTitle = None
        # Invocation Counter.
        self.invocationCounter = 1
        # Used security suite.
        self.securitySuite = SecuritySuite.SUITE_0
        # Signing key pair.
        self.signingKeyPair = None
        # Client key agreement key pair.
        self.keyAgreementKeyPair = None
        # Target (Server or client) Public key.
        self.publicKeys = None
        # Shared secret is generated when connection is made.
        self.sharedSecret = None

    @classmethod
    def decrypt(cls, c, p, data):
        tmp = []
        p.sharedSecret = c.sharedSecret
        tmp = GXDLMSChippering.decryptAesGcm(p, data)
        c.sharedSecret = p.sharedSecret
        return tmp

    #
    # Cipher PDU.
    #      *
    # @param p
    #            Aes GCM Parameter.
    # @param data
    #            Plain text.
    # @return Secured data.
    #
    @classmethod
    def encrypt(cls, p, data):
        if p.security != Security.NONE:
            tmp = GXDLMSChippering.encryptAesGcm(p, data)
            return tmp
        return data

    def reset(self):
        """
        Reset encrypt settings.
        """
        self.security = Security.NONE
        self.invocationCounter = 0

    def isCiphered(self):
        """
        Is ciphering used.
        """
        return self.security != Security.NONE

    #
    # Generate GMAC password from given challenge.
    #      *
    # @param challenge
    #            Client to Server or Server to Client challenge.
    # @return Generated challenge.
    #
    def generateGmacPassword(self, challenge):
        p = AesGcmParameter(
            0x10, self.systemTitle, self.blockCipherKey, self.authenticationKey
        )
        p.security = Security.AUTHENTICATION
        p.invocationCounter = self.invocationCounter
        bb = GXByteBuffer()
        GXDLMSChippering.encryptAesGcm(p, challenge)
        bb.setUInt8(0x10)
        bb.setUInt32(self.invocationCounter)
        bb.set(p.countTag)
        return bb.array()

    def getSecurity(self):
        """
        Used security.
        """
        return self.security

    def setSecurity(self, value):
        """
        Used security.
        """
        self.security = value

    def getSystemTitle(self):
        """
        System title.
        """
        return self.systemTitle

    #
    # Recipient system Title.
    #
    def getRecipientSystemTitle(self):
        """
        Recipient system Title.
        """
        return self.recipientSystemTitle

    def getBlockCipherKey(self):
        # Block cipher key.
        return self.blockCipherKey

    def getAuthenticationKey(self):
        # Authentication key.
        return self.authenticationKey

    def setAuthenticationKey(self, value):
        # Authentication key.
        self.authenticationKey = value

    def getInvocationCounter(self):
        # Invocation counter.
        return self.invocationCounter

    def getSecuritySuite(self):
        # Used security suite.
        return self.securitySuite

    def getEphemeralKeyPair(self):
        # Ephemeral key pair.
        return self.ephemeralKeyPair

    def setEphemeralKeyPair(self, value):
        # Ephemeral key pair.
        self.ephemeralKeyPair = value

    def getKeyAgreementKeyPair(self):
        # Client's key agreement key pair.
        return self.keyAgreementKeyPair

    def setKeyAgreementKeyPair(self, value):
        # Client's key agreement key pair.
        self.keyAgreementKeyPair = value

    def getPublicKeys(self):
        # Target (Server or client) Public key.
        return self.publicKeys

    def getCertificates(self):
        # Available certificates.
        return self.certificates

    def getSigningKeyPair(self):
        # Signing key pair.
        return self.signingKeyPair

    def setSigningKeyPair(self, value):
        # Signing key pair.
        self.signingKeyPair = value

    def getSharedSecret(self):
        # Shared secret is generated when connection is made.
        return self.sharedSecret

    def setSharedSecret(self, value):
        # Shared secret is generated when connection is made.
        self.sharedSecret = value

    def getDedicatedKey(self):
        # Dedicated key.
        return self.dedicatedKey

    def setDedicatedKey(self, value):
        # Dedicated key.
        self.dedicatedKey = value



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------


#
#  * OBIS Code class is used to find out default description to OBIS Code.
#
class GXStandardObisCode:
    # pylint: disable=too-few-public-methods
    #
    # Constructor.
    #
    def __init__(self, forObis, desc=None, forInterfaces=None, forDataType=""):
        # OBIS code.
        if forObis is None:
            self.obis = [0, 0, 0, 0, 0, 0]
        else:
            self.obis = forObis
        # OBIS code description.
        self.description = desc
        # Interfaces that are using this OBIS code.
        self.interfaces = forInterfaces
        # Standard data types.
        self.dataType = forDataType
        # Standard UI data types.
        self.uiDataType = None

    #
    # Convert to string.
    #
    # @return
    #
    def __str__(self):
        str_ = ""
        for s in self.obis:
            if str_:
                str_ += "."
            str_ += s
        return str_ + " " + self.description



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#
#  GXHdlcSettings contains commands for retrieving and setting the limits of
#  field ength and window size, when communicating with the server.
#
class GXHdlcSettings:
    # pylint: disable=too-few-public-methods
    DEFAULT_MAX_INFO_TX = 128
    DEFAULT_MAX_INFO_RX = 128
    DEFAULT_WINDOWS_SIZE_TX = 1
    DEFAULT_WINDOWS_SIZE_RX = 1

    #
    # Constructor.
    #
    def __init__(self):
        self.maxInfoTX = self.DEFAULT_MAX_INFO_TX
        self.maxInfoRX = self.DEFAULT_MAX_INFO_RX
        self.windowSizeTX = self.DEFAULT_WINDOWS_SIZE_TX
        self.windowSizeRX = self.DEFAULT_WINDOWS_SIZE_RX

    # Update HDLC settings from HdlcSetup COSEM object.
    def _update(self, hdlc):
        if hdlc:
            self.maxInfoRX = hdlc.maximumInfoLengthReceive
            self.maxInfoTX = hdlc.maximumInfoLengthTransmit
            self.windowSizeRX = hdlc.windowSizeReceive
            self.windowSizeTX = hdlc.windowSizeTransmit



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
import pkg_resources
from .enums import Standard, ObjectType, DataType
from .GXStandardObisCodeCollection import GXStandardObisCodeCollection
from .GXStandardObisCode import GXStandardObisCode
from .internal._GXCommon import _GXCommon
from .GXEnum import GXEnum
from .GXInt8 import GXInt8
from .GXInt16 import GXInt16
from .GXInt32 import GXInt32
from .GXInt64 import GXInt64
from .GXUInt8 import GXUInt8
from .GXUInt16 import GXUInt16
from .GXUInt32 import GXUInt32
from .GXUInt64 import GXUInt64
from .GXDate import GXDate
from .GXDateTime import GXDateTime
from .GXTime import GXTime
from .GXByteBuffer import GXByteBuffer
from .manufacturersettings.GXObisCode import GXObisCode


###Python 2 requires this
# pylint: disable=bad-option-value,old-style-class
class GXDLMSConverter:
    #
    # Constructor.
    #
    # value: Used standard.
    #
    def __init__(self, value=Standard.DLMS):
        self.standard = value
        # Collection of standard OBIS codes.
        self.codes = GXStandardObisCodeCollection()

    #
    # Get OBIS code description.
    #
    # @param logicalName
    # Logical name (OBIS code).
    # @param type
    # Object type.
    # @param description
    # Description filter.
    # Array of descriptions that match given OBIS code.
    #
    def getDescription(self, logicalName, type_=ObjectType.NONE, description=None):
        if not self.codes:
            self.__readStandardObisInfo(self.standard, self.codes)
        list_ = []
        all_ = not logicalName
        for it in self.codes.find(logicalName, type_):
            if description and not it.description.lower().contains(description.lower()):
                continue
            if all_:
                list_.append(
                    "A="
                    + it.getOBIS()[0]
                    + ", B="
                    + it.getOBIS()[1]
                    + ", C="
                    + it.getOBIS()[2]
                    + ", D="
                    + it.getOBIS()[3]
                    + ", E="
                    + it.getOBIS()[4]
                    + ", F="
                    + it.getOBIS()[5]
                    + "\r\n"
                    + it.description
                )
            else:
                list_.append(it.description)
        return list_

    # pylint: disable=too-many-boolean-expressions
    @classmethod
    def __updateOBISCodeInfo(cls, codes, it, standard):
        ln = it.logicalName
        list_ = codes.find(ln, it.objectType)
        code_ = list_[0]
        if code_:
            if not it.description:
                it.description = code_.description
            # Update data type from DLMS standard.
            if standard != Standard.DLMS:
                d = list_[len(list_) - 1]
                code_.dataType = d.dataType
            if not code_.uiDataType:
                if "10" in code_.dataType:
                    code_.uiDataType = "10"
                elif "25" in code_.dataType or "26" in code_.dataType:
                    code_.uiDataType = "25"
                elif "9" in code_.dataType:
                    if (
                        GXStandardObisCodeCollection.equalsMask2(
                            "0.0-64.96.7.10-14.255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "0.0-64.0.1.5.0-99,255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "0.0-64.0.1.2.0-99,255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.1.2.0-99,255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.1.5.0-99,255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.9.0.255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.9.6.255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.9.7.255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.9.13.255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.9.14.255", ln
                        )
                        or GXStandardObisCodeCollection.equalsMask2(
                            "1.0-64.0.9.15.255", ln
                        )
                    ):
                        code_.uiDataType = "25"
                    # Local time
                    elif GXStandardObisCodeCollection.equalsMask2(
                        "1.0-64.0.9.1.255", ln
                    ):
                        code_.uiDataType = "27"
                    # Local date
                    elif GXStandardObisCodeCollection.equalsMask2(
                        "1.0-64.0.9.2.255", ln
                    ):
                        code_.uiDataType = "26"
                    # Active firmware identifier
                    elif GXStandardObisCodeCollection.equalsMask2("1.0.0.2.0.255", ln):
                        code_.uiDataType = "10"
                # Unix time
                elif (
                    it.objectType == ObjectType.DATA
                    and GXStandardObisCodeCollection.equalsMask2(
                        "0.0.1.1.0.255", it.logicalName
                    )
                ):
                    code_.uiDataType = "25"

            if (
                not code_.dataType == "*"
                and not code_.dataType == ""
                and "," not in code_.dataType
            ):
                tp = int(code_.dataType)
                if it.objectType in (
                    ObjectType.DATA,
                    ObjectType.REGISTER,
                    ObjectType.REGISTER_ACTIVATION,
                    ObjectType.EXTENDED_REGISTER,
                ):
                    it.setDataType(2, tp)
            if code_.uiDataType:
                tp = int(code_.uiDataType)
                if it.objectType in (
                    ObjectType.DATA,
                    ObjectType.REGISTER,
                    ObjectType.REGISTER_ACTIVATION,
                    ObjectType.EXTENDED_REGISTER,
                ):
                    it.setUIDataType(2, tp)
        else:
            print("Unknown OBIS Code: " + it.logicalName + " Type: " + it.objectType)

    def updateOBISCodeInformation(self, objects):
        if not self.codes:
            self.__readStandardObisInfo(self.standard, self.codes)
        if isinstance(objects, list):
            for it in objects:
                self.__updateOBISCodeInfo(self.codes, it, self.standard)
        else:
            self.__updateOBISCodeInfo(self.codes, objects, self.standard)

    @classmethod
    def __getObjects(cls, standard):
        codes = []
        if standard == Standard.ITALY:
            str_ = pkg_resources.resource_string(__name__, "Italy.txt").decode("utf-8")
        elif standard == Standard.INDIA:
            str_ = pkg_resources.resource_string(__name__, "India.txt").decode("utf-8")
        elif standard == Standard.SAUDI_ARABIA:
            str_ = pkg_resources.resource_string(__name__, "SaudiArabia.txt").decode(
                "utf-8"
            )
        elif standard == Standard.SPAIN:
            str_ = pkg_resources.resource_string(__name__, "Spain.txt").decode("utf-8")
        if not str_:
            return None
        str_ = str_.replace("\n", "\r")
        rows = str_.split("\r")
        for it in rows:
            if it and not it.startswith("#"):
                items = it.split(";")
                if len(items) > 1:
                    ot = int(items[0])
                    ln = _GXCommon.toLogicalName(_GXCommon.logicalNameToBytes(items[1]))
                    version = int(items[2])
                    desc = items[3]
                    code_ = GXObisCode(ln, ot, 0, desc)
                    code_.version = version
                    if len(items) > 4:
                        code_.uiDataType = items[4]
                    codes.append(code_)
        return codes

    @classmethod
    def __readStandardObisInfo(cls, standard, codes):
        if standard != Standard.DLMS:
            for it in GXDLMSConverter.__getObjects(standard):
                tmp = GXStandardObisCode(it.logicalName.split("."))
                tmp.interfaces = str(it.objectType)
                tmp.description = it.description
                tmp.uiDataType = it.uiDataType
                codes.append(tmp)

        str_ = pkg_resources.resource_string(__name__, "OBISCodes.txt").decode("utf-8")
        str_ = str_.replace("\n", "\r")
        rows = str_.split("\r")
        for it in rows:
            if it and not it.startswith("#"):
                items = it.split(";")
                obis = items[0].split(".")
                try:
                    code_ = GXStandardObisCode(
                        obis,
                        str(items[3])
                        + "; "
                        + str(items[4])
                        + "; "
                        + str(items[5])
                        + "; "
                        + str(items[6])
                        + "; "
                        + str(items[7]),
                        str(items[1]),
                        str(items[2]),
                    )
                    codes.append(code_)
                except UnicodeEncodeError:
                    pass

    @classmethod
    def changeType(cls, value, type_):
        if _GXCommon.getDLMSDataType(value) == type_:
            return value
        if type_ == DataType.ARRAY:
            raise ValueError("Can't change array types.")
        if type_ == DataType.BCD:
            ret = int(value)
        elif type_ == DataType.BOOLEAN:
            ret = bool(int(value))
        elif type_ == DataType.COMPACT_ARRAY:
            raise ValueError("Can't change compact array types.")
        elif type_ == DataType.DATE:
            ret = GXDate(value)
        elif type_ == DataType.DATETIME:
            ret = GXDateTime(value)
        elif type_ == DataType.ENUM:
            ret = GXEnum(value)
        elif type_ == DataType.FLOAT32:
            ret = float(value)
        elif type_ == DataType.FLOAT64:
            ret = float(value)
        elif type_ == DataType.INT16:
            ret = GXInt16(value)
        elif type_ == DataType.INT32:
            ret = GXInt32(value)
        elif type_ == DataType.INT64:
            ret = GXInt64(value)
        elif type_ == DataType.INT8:
            ret = GXInt8(value)
        elif type_ == DataType.NONE:
            ret = None
        elif type_ == DataType.OCTET_STRING:
            if isinstance(value, str):
                ret = GXByteBuffer.hexToBytes(str(value))
            else:
                raise ValueError("Can't change octect string type.")
        elif type_ == DataType.STRING:
            ret = str(value)
        elif type_ == DataType.BITSTRING:
            ret = str(value)
        elif type_ == DataType.STRING_UTF8:
            ret = str(value)
        elif type_ == DataType.STRUCTURE:
            raise ValueError("Can't change structure types.")
        elif type_ == DataType.TIME:
            ret = GXTime(value)
        elif type_ == DataType.UINT16:
            ret = GXUInt16(value)
        elif type_ == DataType.UINT32:
            ret = GXUInt32(value)
        elif type_ == DataType.UINT64:
            ret = GXUInt64(value)
        elif type_ == DataType.UINT8:
            ret = GXUInt8(value)
        else:
            raise ValueError("Invalid data type.")
        return ret

    @classmethod
    def objectTypeToString(cls, ot):
        if ot == ObjectType.ACTION_SCHEDULE:
            ret = "ActionSchedule"
        elif ot == ObjectType.ACTIVITY_CALENDAR:
            ret = "ActivityCalendar"
        elif ot == ObjectType.ASSOCIATION_LOGICAL_NAME:
            ret = "AssociationLogicalName"
        elif ot == ObjectType.ASSOCIATION_SHORT_NAME:
            ret = "AssociationShortName"
        elif ot == ObjectType.AUTO_ANSWER:
            ret = "AutoAnswer"
        elif ot == ObjectType.AUTO_CONNECT:
            ret = "AutoConnect"
        elif ot == ObjectType.CLOCK:
            ret = "Clock"
        elif ot == ObjectType.DATA:
            ret = "Data"
        elif ot == ObjectType.DEMAND_REGISTER:
            ret = "DemandRegister"
        elif ot == ObjectType.MAC_ADDRESS_SETUP:
            ret = "MacAddressSetup"
        elif ot == ObjectType.EXTENDED_REGISTER:
            ret = "ExtendedRegister"
        elif ot == ObjectType.GPRS_SETUP:
            ret = "GprsSetup"
        elif ot == ObjectType.SECURITY_SETUP:
            ret = "SecuritySetup"
        elif ot == ObjectType.IEC_HDLC_SETUP:
            ret = "IecHdlcSetup"
        elif ot == ObjectType.IEC_LOCAL_PORT_SETUP:
            ret = "IecLocalPortSetup"
        elif ot == ObjectType.IEC_TWISTED_PAIR_SETUP:
            ret = "IEC_TWISTED_PAIR_SETUP"
        elif ot == ObjectType.IP4_SETUP:
            ret = "Ip4Setup"
        elif ot == ObjectType.MBUS_SLAVE_PORT_SETUP:
            ret = "MBusSlavePortSetup"
        elif ot == ObjectType.IMAGE_TRANSFER:
            ret = "ImageTransfer"
        elif ot == ObjectType.DISCONNECT_CONTROL:
            ret = "DisconnectControl"
        elif ot == ObjectType.LIMITER:
            ret = "Limiter"
        elif ot == ObjectType.MBUS_CLIENT:
            ret = "MBusClient"
        elif ot == ObjectType.MODEM_CONFIGURATION:
            ret = "ModemConfiguration"
        elif ot == ObjectType.PPP_SETUP:
            ret = "PppSetup"
        elif ot == ObjectType.PROFILE_GENERIC:
            ret = "ProfileGeneric"
        elif ot == ObjectType.REGISTER:
            ret = "Register"
        elif ot == ObjectType.REGISTER_ACTIVATION:
            ret = "RegisterActivation"
        elif ot == ObjectType.REGISTER_MONITOR:
            ret = "RegisterMonitor"
        elif ot == ObjectType.REGISTER_TABLE:
            ret = "RegisterTable"
        elif ot == ObjectType.ZIG_BEE_SAS_STARTUP:
            ret = "ZigBeeSasStartup"
        elif ot == ObjectType.ZIG_BEE_SAS_JOIN:
            ret = "ZigBeeSasJoin"
        elif ot == ObjectType.ZIG_BEE_SAS_APS_FRAGMENTATION:
            ret = "ZigBeeSasApsFragmentation"
        elif ot == ObjectType.ZIG_BEE_NETWORK_CONTROL:
            ret = "ZigBeeNetworkControl"
        elif ot == ObjectType.SAP_ASSIGNMENT:
            ret = "SapAssignment"
        elif ot == ObjectType.SCHEDULE:
            ret = "Schedule"
        elif ot == ObjectType.SCRIPT_TABLE:
            ret = "ScriptTable"
        elif ot == ObjectType.SMTP_SETUP:
            ret = "SMTPSetup"
        elif ot == ObjectType.SPECIAL_DAYS_TABLE:
            ret = "SpecialDaysTable"
        elif ot == ObjectType.STATUS_MAPPING:
            ret = "StatusMapping"
        elif ot == ObjectType.TCP_UDP_SETUP:
            ret = "TcpUdpSetup"
        elif ot == ObjectType.UTILITY_TABLES:
            ret = "UtilityTables"
        elif ot == ObjectType.MBUS_MASTER_PORT_SETUP:
            ret = "MBusMasterPortSetup"
        elif ot == ObjectType.PUSH_SETUP:
            ret = "PushSetup"
        elif ot == ObjectType.ACCOUNT:
            ret = "Account"
        elif ot == ObjectType.CREDIT:
            ret = "Credit"
        elif ot == ObjectType.CHARGE:
            ret = "Charge"
        elif ot == ObjectType.PARAMETER_MONITOR:
            ret = "ParameterMonitor"
        elif ot == ObjectType.TOKEN_GATEWAY:
            ret = "Token"
        elif ot == ObjectType.GSM_DIAGNOSTIC:
            ret = "GSMDiagnostic"
        elif ot == ObjectType.COMPACT_DATA:
            ret = "CompactData"
        elif ot == ObjectType.IP6_SETUP:
            ret = "Ip6Setup"
        elif ot == ObjectType.LLC_SSCS_SETUP:
            ret = "LlcSscsSetup"
        elif ot == ObjectType.PRIME_NB_OFDM_PLC_PHYSICAL_LAYER_COUNTERS:
            ret = "PrimeNbOfdmPlcPhysicalLayerCounters"
        elif ot == ObjectType.PRIME_NB_OFDM_PLC_MAC_SETUP:
            ret = "PrimeNbOfdmPlcMacSetup"
        elif ot == ObjectType.PRIME_NB_OFDM_PLC_MAC_FUNCTIONAL_PARAMETERS:
            ret = "PrimeNbOfdmPlcMacFunctionalParameters"
        elif ot == ObjectType.PRIME_NB_OFDM_PLC_MAC_COUNTERS:
            ret = "PrimeNbOfdmPlcMacCounters"
        elif ot == ObjectType.PRIME_NB_OFDM_PLC_MAC_NETWORK_ADMINISTRATION_DATA:
            ret = "PrimeNbOfdmPlcMacNetworkAdministrationData"
        elif ot == ObjectType.PRIME_NB_OFDM_PLC_APPLICATIONS_IDENTIFICATION:
            ret = "PrimeNbOfdmPlcApplicationsIdentification"
        elif ot == ObjectType.NTP_SETUP:
            ret = "NtpSetup"
        elif ot == ObjectType.FUNCTION_CONTROL:
            ret = "FunctionControl"
        else:
            ret = "Manufacture spesific."
        return ret

    @classmethod
    def valueOfObjectType(cls, value):
        if value == "ActionSchedule":
            ot = ObjectType.ACTION_SCHEDULE
        elif value == "ActivityCalendar":
            ot = ObjectType.ACTIVITY_CALENDAR
        elif value == "AssociationLogicalName":
            ot = ObjectType.ASSOCIATION_LOGICAL_NAME
        elif value == "AssociationShortName":
            ot = ObjectType.ASSOCIATION_SHORT_NAME
        elif value == "AutoAnswer":
            ot = ObjectType.AUTO_ANSWER
        elif value == "AutoConnect":
            ot = ObjectType.AUTO_CONNECT
        elif value == "Clock":
            ot = ObjectType.CLOCK
        elif value == "Data":
            ot = ObjectType.DATA
        elif value == "DemandRegister":
            ot = ObjectType.DEMAND_REGISTER
        elif value == "MacAddressSetup":
            ot = ObjectType.MAC_ADDRESS_SETUP
        elif value == "ExtendedRegister":
            ot = ObjectType.EXTENDED_REGISTER
        elif value == "GprsSetup":
            ot = ObjectType.GPRS_SETUP
        elif value == "SecuritySetup":
            ot = ObjectType.SECURITY_SETUP
        elif value in ("IecHdlcSetup", "HdlcSetup"):
            ot = ObjectType.IEC_HDLC_SETUP
        elif value in ("IecLocalPortSetup", "IecOpticalPortSetup"):
            ot = ObjectType.IEC_LOCAL_PORT_SETUP
        elif value == "IEC_TWISTED_PAIR_SETUP":
            ot = ObjectType.IEC_TWISTED_PAIR_SETUP
        elif value == "Ip4Setup":
            ot = ObjectType.IP4_SETUP
        elif value == "MBusSlavePortSetup":
            ot = ObjectType.MBUS_SLAVE_PORT_SETUP
        elif value == "ImageTransfer":
            ot = ObjectType.IMAGE_TRANSFER
        elif value == "DisconnectControl":
            ot = ObjectType.DISCONNECT_CONTROL
        elif value == "Limiter":
            ot = ObjectType.LIMITER
        elif value == "MBusClient":
            ot = ObjectType.MBUS_CLIENT
        elif value == "ModemConfiguration":
            ot = ObjectType.MODEM_CONFIGURATION
        elif value == "PppSetup":
            ot = ObjectType.PPP_SETUP
        elif value == "ProfileGeneric":
            ot = ObjectType.PROFILE_GENERIC
        elif value == "Register":
            ot = ObjectType.REGISTER
        elif value == "RegisterActivation":
            ot = ObjectType.REGISTER_ACTIVATION
        elif value == "RegisterMonitor":
            ot = ObjectType.REGISTER_MONITOR
        elif value == "RegisterTable":
            ot = ObjectType.REGISTER_TABLE
        elif value == "ZigBeeSasStartup":
            ot = ObjectType.ZIG_BEE_SAS_STARTUP
        elif value == "ZigBeeSasJoin":
            ot = ObjectType.ZIG_BEE_SAS_JOIN
        elif value == "ZigBeeSasApsFragmentation":
            ot = ObjectType.ZIG_BEE_SAS_APS_FRAGMENTATION
        elif value == "ZigBeeNetworkControl":
            ot = ObjectType.ZIG_BEE_NETWORK_CONTROL
        elif value == "SapAssignment":
            ot = ObjectType.SAP_ASSIGNMENT
        elif value == "Schedule":
            ot = ObjectType.SCHEDULE
        elif value == "ScriptTable":
            ot = ObjectType.SCRIPT_TABLE
        elif value == "SMTPSetup":
            ot = ObjectType.SMTP_SETUP
        elif value == "SpecialDaysTable":
            ot = ObjectType.SPECIAL_DAYS_TABLE
        elif value == "StatusMapping":
            ot = ObjectType.STATUS_MAPPING
        elif value == "TcpUdpSetup":
            ot = ObjectType.TCP_UDP_SETUP
        elif value == "UtilityTables":
            ot = ObjectType.UTILITY_TABLES
        elif value == "MBusMasterPortSetup":
            ot = ObjectType.MBUS_MASTER_PORT_SETUP
        elif value == "PushSetup":
            ot = ObjectType.PUSH_SETUP
        elif value == "Account":
            ot = ObjectType.ACCOUNT
        elif value == "Credit":
            ot = ObjectType.CREDIT
        elif value == "Charge":
            ot = ObjectType.CHARGE
        elif value == "ParameterMonitor":
            ot = ObjectType.PARAMETER_MONITOR
        elif value in ("Token", "TokenGateway"):
            ot = ObjectType.TOKEN_GATEWAY
        elif value == "GSMDiagnostic":
            ot = ObjectType.GSM_DIAGNOSTIC
        elif value == "CompactData":
            ot = ObjectType.COMPACT_DATA
        elif value == "Ip6Setup":
            ot = ObjectType.IP6_SETUP
        elif value == "LlcSscsSetup":
            ot = ObjectType.LLC_SSCS_SETUP
        elif value == "PrimeNbOfdmPlcPhysicalLayerCounters":
            ot = ObjectType.PRIME_NB_OFDM_PLC_PHYSICAL_LAYER_COUNTERS
        elif value == "PrimeNbOfdmPlcMacSetup":
            ot = ObjectType.PRIME_NB_OFDM_PLC_MAC_SETUP
        elif value == "PrimeNbOfdmPlcMacFunctionalParameters":
            ot = ObjectType.PRIME_NB_OFDM_PLC_MAC_FUNCTIONAL_PARAMETERS
        elif value == "PrimeNbOfdmPlcMacCounters":
            ot = ObjectType.PRIME_NB_OFDM_PLC_MAC_COUNTERS
        elif value == "PrimeNbOfdmPlcMacNetworkAdministrationData":
            ot = ObjectType.PRIME_NB_OFDM_PLC_MAC_NETWORK_ADMINISTRATION_DATA
        elif value == "PrimeNbOfdmPlcApplicationsIdentification":
            ot = ObjectType.PRIME_NB_OFDM_PLC_APPLICATIONS_IDENTIFICATION
        elif value == "NtpSetup":
            ot = ObjectType.NTP_SETUP
        elif value == "FunctionControl":
            ot = ObjectType.FUNCTION_CONTROL
        else:
            ot = ObjectType.NONE
        return ot
